#include "BlockManager.h"
#include <assert.h>
#include <math.h>

#include <RendererManager.h>
#include <Transformer.h>
#include <Scene.h>
#include <StringTools.h>

#include "../SurfelGenerator/SurfelGenerator.h"

static const double BLOCK_MIN_SIZE = 0.01; ///< minimum authorized
static const unsigned int BLOCK_SEQ_LEN = 4; ///< number of digit for block number

static BlockManager * instance = NULL;

BlockManager * BlockManager::getInstance () {
  if (instance == NULL)
    instance = new BlockManager;
  return instance;
}
  
void BlockManager::destroyInstance () {
  if (instance != NULL) {
    delete instance;
    instance = NULL;
  }
}    

BlockManager::BlockManager () : mousePressed (false), blockVisible (false), 
				currentPrimitive (NULL), currentPrimitiveObject (NULL),
				instanceCount (0) {
}

BlockManager::~BlockManager () { 
  clean ();
}

PrimitiveInterface * BlockManager::getCurrentPrimitive () {
  PrimitiveInterface * res = currentPrimitive;
 
  currentPrimitive = NULL;
  return res;
} 

Object * BlockManager::getCurrentObject () {
  Object * res = currentPrimitiveObject;
  
  currentPrimitiveObject = NULL;
  return res;
}

PrimitiveInterface * BlockManager::createPrimitive () {
  QString newName = StringTools::appendSequenceNumber ("Block", 
						       instanceCount++, 
						       BLOCK_SEQ_LEN);
  return new Block (newName);
} 

void BlockManager::clean () {
  if (currentPrimitive != NULL)
    delete currentPrimitive;
  if (currentPrimitiveObject != NULL)
    delete currentPrimitiveObject;
}


void BlockManager::handleMousePressEvent (QMouseEvent *event, 
					   CoTeXConfigurationWidget * confWidget) {
  assert (confWidget != NULL);
  
  if (!mousePressed 
      && event->button () == Qt::LeftButton 
      && event->state () == Qt::NoButton) {
    SurfelInterface * surfel = NULL;
    
    clean ();
    mousePressed = true;
    configurationWidget = confWidget;
    originalMousePosition = event->pos();
    lastMousePosition = originalMousePosition;
    Scene * scene = Scene::getInstance ();
    Object * activeObject = scene->getActiveObject ();
    RendererInterface * renderer = RendererManager::getInstance()->getMainRenderer ();
    renderer->renderFrame (FrameBufferInterface::PERPIXEL_SURFEL 
			   | FrameBufferInterface::PERPIXEL_C_Z_N_W);
    FrameBufferInterface * frameBuffer = renderer->getFrameBuffer();
    if (frameBuffer->getObjectAtPixel (event->x (), event->y ()) != activeObject) {
      mousePressed = false;
      return; // only surfels from active object are interesting
    }
    surfel = frameBuffer->getSurfel (event->x (), event->y ());
    if (surfel != NULL) {
      center = surfel->getPosition ();
      screenSize = 0.0;
    }
    else
      mousePressed = false;
  }
}


void BlockManager::handleMouseMoveEvent (QMouseEvent * event) {
  
  FrameBufferInterface * frameBuffer;
  RendererInterface * renderer;
  Transformer * transformer;
  
  if (mousePressed
      && event->button() == Qt::NoButton) {
    if (blockVisible)
      drawBlock (); // for erasing the last circle (inverted colors).
    renderer = RendererManager::getInstance ()->getMainRenderer (); // resamplingRenderer
    frameBuffer = renderer->getFrameBuffer ();
    float depth = frameBuffer->getDepth (originalMousePosition.x (), 
					 originalMousePosition.y ());
      
    transformer = Transformer::getSceneTransformer(); // Returns a global instance of Transformer,which is updated by the RendererManager
    Vector3D point1 = transformer->viewPortToObject (Vector2D (originalMousePosition.x (),
						      originalMousePosition.y ()),
					    depth);
    Vector3D point2 = transformer->viewPortToObject (Vector2D ( lastMousePosition.x (),  
						       lastMousePosition.y ()),
					    depth);
    distance = Vector3D::distance (point1, point2);
    
    lastMousePosition = event->pos ();
    QPoint diff = event->pos () - originalMousePosition;
    screenSize = sqrt (diff.x () * diff.x () + diff.y () * diff.y ());
    
    drawBlock ();
    blockVisible = true;
  }
}


void BlockManager::handleMouseReleaseEvent (QMouseEvent *) {
  if (mousePressed) {
    mousePressed = false;
    if (originalMousePosition == lastMousePosition) { 
      delete currentPrimitive;
      currentPrimitive = NULL;
      throw PrimitiveManagerException ("Too Small Size");
    }
    else {
      Scene * scene = Scene::getInstance ();
      Object * activeObject = scene->getActiveObject ();
      RendererInterface * renderer = RendererManager::getInstance ()->getMainRenderer ();
      FrameBufferInterface * frameBuffer = renderer->getFrameBuffer ();;

      Transformer * transformer;
      Vector3D point1, point2;

      float depth = frameBuffer->getDepth (originalMousePosition.x (), 
					   originalMousePosition.y ());
      
      transformer = Transformer::getSceneTransformer(); // Returns a global instance of Transformer,which is updated by the RendererManager
      point1 = transformer->viewPortToObject (Vector2D (originalMousePosition.x (),
							originalMousePosition.y ()),
					      depth);
      point2 = transformer->viewPortToObject (Vector2D ( lastMousePosition.x (),  
							 lastMousePosition.y ()),
					      depth);
      distance = Vector3D::distance (point1, point2);
      // unique name  : "Block0001", "Block0002", ...
      QString newName = StringTools::appendSequenceNumber ("Block", 
							   instanceCount++, 
							   BLOCK_SEQ_LEN);

      center -= Vector3D (distance, distance, distance); // for having the block center on the mouse
		  

      currentPrimitive = new Block (newName, center,
				    2 * distance, 2 * distance, 2 * distance, 
				    configurationWidget->getCurrentTexture ());

      
      currentPrimitiveObject = new Object (scene->getUniqueName ("CoTeXBlock"));
      
      // PrimitivesObjects, contrary to primitives, must be expressed in the 
      // global mark.
      currentPrimitiveObject->setPosition (activeObject->getPosition ()); 
					   
      
      scene->addObject (currentPrimitiveObject, false, false);

      ColorAttribute primitiveObjectColor = 
	createColorForViz (currentPrimitive->getTexture (),
			   currentPrimitive->getCenter ());
      BBoxMinMax bBoxMinMax = currentPrimitive->getBBoxMinMax ();
      
      generatePrimitiveSurfels (currentPrimitive, currentPrimitiveObject, bBoxMinMax,
				primitiveObjectColor);
      blockVisible = false; // scene re-rendered
    }
  }
}


void BlockManager::handleKeyPressEvent (QKeyEvent *ke) {
  ke->ignore(); // by now, we'll ignore the key event
}


void BlockManager::handleKeyReleaseEvent (QKeyEvent *ke) {
  ke->ignore(); // by now, we'll ignore the key event
}

// ***************************************
// VISUALISATION OF THE BLOCK BY A SEGMENT
// ***************************************

void BlockManager::drawBlock () {
  
  Transformer * transformer = Transformer::getSceneTransformer ();
  
  p3D[0] = center + Vector3D (-distance, -distance, distance); 
  p3D[1] = center + Vector3D (distance, -distance, distance); 
  p3D[2] = center + Vector3D (distance, -distance, -distance); 
  p3D[3] = center + Vector3D (-distance, -distance, -distance); 
  p3D[4] = center + Vector3D (-distance, distance, distance); 
  p3D[5] = center + Vector3D (distance, distance, distance); 
  p3D[6] = center + Vector3D (distance, distance, -distance); 
  p3D[7] = center + Vector3D (-distance, distance, -distance); 
  
  int i;
  for (i = 0; i < 8; i++)
    p2D[i] = transformer->objectToViewPort (p3D[i]);
  
  
  // Draw a line representing the size of the block
  painter.begin (RendererManager::getInstance()->getMainRenderer()->getRendererWidget());
  
  // initialisation of the pen for having a good visualization in all cases 
  // (inverted color).
  painter.setPen (QPen (Qt::white, 1, Qt::SolidLine));
  painter.setRasterOp (Qt::XorROP);
  
  // Draw lines for a block
  painter.drawLine ((int)p2D[0][0], (int)p2D[0][1], (int)p2D[1][0], (int)p2D[1][1]);
  painter.drawLine ((int)p2D[1][0], (int)p2D[1][1], (int)p2D[2][0], (int)p2D[2][1]);
  painter.drawLine ((int)p2D[2][0], (int)p2D[2][1], (int)p2D[3][0], (int)p2D[3][1]);
  painter.drawLine ((int)p2D[3][0], (int)p2D[3][1], (int)p2D[0][0], (int)p2D[0][1]);
  painter.drawLine ((int)p2D[4][0], (int)p2D[4][1], (int)p2D[5][0], (int)p2D[5][1]);
  painter.drawLine ((int)p2D[5][0], (int)p2D[5][1], (int)p2D[6][0], (int)p2D[6][1]);
  painter.drawLine ((int)p2D[6][0], (int)p2D[6][1], (int)p2D[7][0], (int)p2D[7][1]);
  painter.drawLine ((int)p2D[7][0], (int)p2D[7][1], (int)p2D[4][0], (int)p2D[4][1]);
  painter.drawLine ((int)p2D[0][0], (int)p2D[0][1], (int)p2D[4][0], (int)p2D[4][1]);
  painter.drawLine ((int)p2D[1][0], (int)p2D[1][1], (int)p2D[5][0], (int)p2D[5][1]);
  painter.drawLine ((int)p2D[2][0], (int)p2D[2][1], (int)p2D[6][0], (int)p2D[6][1]);
  painter.drawLine ((int)p2D[3][0], (int)p2D[3][1], (int)p2D[7][0], (int)p2D[7][1]);
  
  painter.flush ();
  painter.end ();
}
