#include "ConvPointManager.h"
#include <assert.h>
#include <qpainter.h>
#include <RendererManager.h>
#include <Scene.h>
#include <StringTools.h>

using namespace std;

static const unsigned int CONVCURVE_SEQ_LEN = 4; ///< number of digit for sphere number
static const float CONVCURVE_BASE_DISTANCE = 10.0; ///< the minimun base distance between 2 skeleton points 

static ConvPointManager * instance = NULL;


ConvPointManager * ConvPointManager::getInstance () {
  if (instance == NULL)
    instance = new ConvPointManager;
  return instance;
}
  
void ConvPointManager::destroyInstance () {
  if (instance != NULL) {
    delete instance;
    instance = NULL;
  }
}    

ConvPointManager::ConvPointManager () : mousePressed (false), skeleton (NULL),
					currentPrimitive (NULL), currentPrimitiveObject (NULL),
					instanceCount (0) {
}


ConvPointManager::~ConvPointManager () { 
  clean ();
}

PrimitiveInterface* ConvPointManager::getCurrentPrimitive () {
  PrimitiveInterface * res = currentPrimitive;
  
  currentPrimitive = NULL;
  return res;
}

Object * ConvPointManager::getCurrentObject () {
  Object * res = currentPrimitiveObject;
  
  currentPrimitiveObject = NULL;
  return res;
}

PrimitiveInterface * ConvPointManager::createPrimitive () {
  QString newName = StringTools::appendSequenceNumber ("ConvPoint", // for unicity
						       instanceCount++, 
						       CONVCURVE_SEQ_LEN);
  ConvPoint::ConvSkeleton * tempSke = new ConvPoint::ConvSkeleton();
  tempSke->resize (3); /// 3 point at (0,0,0)
  return new ConvPoint (newName, 1.0, tempSke); // S to 1.0
}

void ConvPointManager::clean () {
  
  if (currentPrimitive != NULL) {
    delete currentPrimitive;
    currentPrimitive = NULL;
  }
  
  if (skeleton != NULL) {
    delete skeleton;
  }
  
  if (currentPrimitiveObject != NULL) {
    delete currentPrimitiveObject;
    currentPrimitiveObject = NULL;
  }
}


void ConvPointManager::handleMousePressEvent (QMouseEvent *event, 
					      CoTeXConfigurationWidget * confWidget) {
  assert (confWidget != NULL);
  
  if (!mousePressed 
      && event->button() == Qt::LeftButton 
      && event->state() == Qt::NoButton) {
    SurfelInterface * surfel = NULL;
    Vector3D center;
    
    clean ();
    mousePressed = true;
    configurationWidget = confWidget;
    skeleton = new ConvPoint::ConvSkeleton;
    RendererInterface * renderer = RendererManager::getInstance()->getMainRenderer();
    renderer->renderFrame (FrameBufferInterface::PERPIXEL_SURFEL 
			   | FrameBufferInterface::PERPIXEL_C_Z_N_W);
    FrameBufferInterface * frameBuffer = renderer->getFrameBuffer();
    Scene * scene = Scene::getInstance ();
    Object * activeObject = scene->getActiveObject ();
    if (frameBuffer->getObjectAtPixel (event->x(), event->y()) != activeObject) {
      mousePressed = false;
      return; // only surfels from active object are interesting
    }
    surfel = frameBuffer->getSurfel (event->x(), event->y());
    if (surfel != NULL) {
      Vector3D tmp = surfel->getPosition () - activeObject->getPosition ();
      skeleton->push_back (Vector3D (tmp[0], tmp[1], tmp[2]));
    }
    else
      mousePressed = false;
  }
}


void ConvPointManager::handleMouseMoveEvent (QMouseEvent *event) {
  if (mousePressed) {
    Object * activeObject = Scene::getInstance ()->getActiveObject ();
    FrameBufferInterface * frameBuffer = 
      RendererManager::getInstance()->getMainRenderer()->getFrameBuffer ();
    SurfelInterface * surfel = NULL;
    surfel = frameBuffer->getSurfel (event->x(), event->y());
    
    if (surfel != NULL
	&& frameBuffer->getObjectAtPixel (event->x(), event->y()) == activeObject) {
      
      // ********************************
      // Draw the skeleton on the screen.
      // ********************************
      QPainter painter;
      painter.begin (RendererManager::getInstance()->getMainRenderer()->getRendererWidget());
      painter.setPen (QPen (Qt::white, 1, Qt::SolidLine));
      painter.setRasterOp (Qt::XorROP);
      painter.drawPoint (event->x (), event->y ());
      painter.flush ();
      painter.end ();
      
      // ****************************************************
      // Only few points of the skeleton are realy important.
      // Here we decide according to S, if the last point
      // was enough far from the current for pushing back the 
      // current
      // ****************************************************
      
      float currentS = configurationWidget->getConvS ();
      float currentDistance = (surfel->getPosition () - 
			       skeleton->back ()).getLength ();
      if (currentDistance > CONVCURVE_BASE_DISTANCE / currentS) {
	Vector3D newPoint = surfel->getPosition ();
	skeleton->push_back (newPoint);    
      }
    }
  }
}


void ConvPointManager::handleMouseReleaseEvent (QMouseEvent *) {
  if (mousePressed) {
    mousePressed = false;
    if (skeleton->size () >= 2) {  // at least 2 point for a convolution
      
      Scene * scene = Scene::getInstance ();
      Object * activeObject = scene->getActiveObject (); 
      QString newName = StringTools::appendSequenceNumber ("ConvPoint", 
							   instanceCount++, 
							   CONVCURVE_SEQ_LEN);
      TextureInterface * texture = configurationWidget->getCurrentTexture ();
      double s = configurationWidget->getConvS ();
      double threshold = configurationWidget->getConvThreshold ();
      
      currentPrimitive = new ConvPoint (newName, s, skeleton, texture, threshold);
      currentPrimitiveObject = new Object (scene->getUniqueName ("CoTeXConvPoint"));
      currentPrimitiveObject->setPosition (activeObject->getPosition ());
      scene->addObject (currentPrimitiveObject, false, false);
      
      // generating a new color, different from the texture, for better viz.
      ColorAttribute primitiveObjectColor = createColorForViz (texture,
							       skeleton->front ());
      
      BBoxMinMax bBoxMinMax = currentPrimitive->getBBoxMinMax ();
      generatePrimitiveSurfels (currentPrimitive, currentPrimitiveObject, bBoxMinMax,
				primitiveObjectColor);
      
      skeleton = NULL;
    }
    else {
      delete skeleton;
      skeleton = NULL;
      throw PrimitiveManagerException ("Too small Convolution");
    }      
  }
}


void ConvPointManager::handleKeyPressEvent (QKeyEvent * ke) {
  ke->ignore(); // by now, we'll ignore the key event
}


void ConvPointManager::handleKeyReleaseEvent (QKeyEvent * ke) {
  ke->ignore(); // by now, we'll ignore the key event
}
