#include "RbfManager.h"

#include <assert.h>
#include <qpainter.h>

#include <Transformer.h>
#include <RendererManager.h>
#include <Scene.h>
#include <StringTools.h>
#include <FrameBufferInterface.h>

#include "../SurfelGenerator/SurfelGenerator.h"

static RbfManager * instance = NULL;
static Object * object = NULL;

RbfManager * RbfManager::getInstance () {
  if (instance == NULL)
    instance = new RbfManager;
  return instance;
}
  
void RbfManager::destroyInstance () {
  if (instance != NULL) {
    delete instance;
    instance = NULL;
  }
} 

RbfManager::RbfManager () : mousePressed (false), rbfPointSet (NULL), 
			    currentPrimitive (NULL), currentPrimitiveObject (NULL),
			    instanceCount (0) {
}

RbfManager::~RbfManager () { 
  clean ();
}

PrimitiveInterface* RbfManager::getCurrentPrimitive () {
  PrimitiveInterface * res = currentPrimitive;
 
  currentPrimitive = NULL;
  return res;
}

Object * RbfManager::getCurrentObject () {
  Object * res = currentPrimitiveObject;
  
  currentPrimitiveObject = NULL;
  return res;
}

PrimitiveInterface * RbfManager::createPrimitive () {
  QString newName = StringTools::appendSequenceNumber ("Rbf", 
						       instanceCount++, 
						       RBF_SEQ_LEN);
  Rbf::RbfSet * tempSet = new Rbf::RbfSet ();
  tempSet->resize (RBF_MIN_POINTSET_SIZE); // minimum number of points for RBF
  (*tempSet)[0] = Vector3D (0,0,0);
  (*tempSet)[1] = Vector3D (0,0,0);
  (*tempSet)[2] = Vector3D (0,0,0);
  return new Rbf (newName, tempSet);
}

void RbfManager::clean () {
  if (currentPrimitive != NULL) {
    delete currentPrimitive;
    currentPrimitive = NULL;
  }
  if (rbfPointSet != NULL)
    delete rbfPointSet;
  if (currentPrimitiveObject != NULL) {
    delete currentPrimitiveObject;
    currentPrimitiveObject = NULL;
  }
  object = NULL;
}

void RbfManager::drawRbfPoints () {
  assert (rbfPointSet != NULL);  
  
  Transformer * transformer = Transformer::getSceneTransformer();
  QPainter painter;
  
  painter.begin (RendererManager::getInstance()->getMainRenderer()->getRendererWidget());
  
  unsigned int i;
  
  // *****************************************
  // Draw visibles 2D points for visualisation
  // *****************************************

  for (i = 0; i < rbfPointSet->size (); i++) {
    Vector3D pos = (*rbfPointSet)[i];
    Vector2D point2D = transformer->objectToViewPort (pos);
    
    painter.setPen (QPen (Qt::white, 1, Qt::SolidLine));
    painter.drawPoint ((int)point2D[0],(int)point2D[1]);
    painter.setPen (QPen (Qt::red, 1, Qt::SolidLine));
    painter.drawPoint ((int)point2D[0] + 1,(int)point2D[1] + 1);
    painter.drawPoint ((int)point2D[0] + 1, (int)point2D[1] - 1);
    painter.drawPoint ((int)point2D[0] - 1, (int)point2D[1] + 1);
    painter.drawPoint ((int)point2D[0] - 1, (int)point2D[1] - 1);
    painter.drawPoint ((int)point2D[0], (int)point2D[1] + 1);
    painter.drawPoint ((int)point2D[0], (int)point2D[1] - 1);
    painter.drawPoint ((int)point2D[0] + 1, (int)point2D[1]);
    painter.drawPoint ((int)point2D[0] - 1, (int)point2D[1]);
  }  
  painter.flush ();
  painter.end ();

}


void RbfManager::handleMousePressEvent (QMouseEvent *event, 
					CoTeXConfigurationWidget * confWidget) {
  SurfelInterface *surfel;
  Scene * scene = Scene::getInstance ();
  Object *activeObject = scene->getActiveObject ();
  if (event->button () == Qt::LeftButton || event->button () == Qt::RightButton){
    if (!mousePressed || activeObject != object){///The first time we pressed the mouse, we initialize everything
      clean ();
      rbfPointSet = new Rbf::RbfSet;
      mousePressed = true;
      configurationWidget = confWidget;
    }
    object = activeObject;
    surfel = NULL;
    RendererInterface * renderer = RendererManager::getInstance()->getMainRenderer();
    renderer->renderFrame (FrameBufferInterface::PERPIXEL_SURFEL 
			   | FrameBufferInterface::PERPIXEL_C_Z_N_W);
    
    FrameBufferInterface *frameBuffer = renderer->getFrameBuffer ();
    if (frameBuffer->getObjectAtPixel (event->x (), event->y ()) != activeObject) {
      mousePressed = false;
      renderer->renderFrame ();
      throw PrimitiveManagerException ("Point not valid"); 
    }
    surfel = frameBuffer->getSurfel (event->x (), event->y ());    
    if (surfel != NULL) {
      Vector3D pos = surfel->getPosition ();
      rbfPointSet->push_back (pos);
    }
    else
      throw PrimitiveManagerException ("Point not valid"); 
    drawRbfPoints ();
  }
}

void RbfManager::handleMouseReleaseEvent (QMouseEvent *event) {
  Scene *scene = Scene::getInstance ();
  Object *activeObject = scene->getActiveObject ();
  if (event->state() == Qt::RightButton && mousePressed) {
    mousePressed = false; 
    if (rbfPointSet->size() >= RBF_MIN_POINTSET_SIZE
	&& rbfPointSet->size() <= RBF_MAX_POINTSET_SIZE){
      TextureInterface * texture = configurationWidget->getCurrentTexture ();
      QString newName = StringTools::appendSequenceNumber ("Rbf", 
							   instanceCount++, 
							   RBF_SEQ_LEN);
      currentPrimitive = new Rbf (newName, rbfPointSet, texture);
      currentPrimitiveObject = new Object (scene->getUniqueName ("CoTeXRbf"));
      currentPrimitiveObject->setPosition (activeObject->getPosition ());
      scene->addObject (currentPrimitiveObject, false, false);
      
      ColorAttribute primitiveObjectColor = createColorForViz (texture,
							       rbfPointSet->front ());
      BBoxMinMax bBoxMinMax = currentPrimitive->getBBoxMinMax ();
      generatePrimitiveSurfels (currentPrimitive, currentPrimitiveObject, bBoxMinMax,
				primitiveObjectColor);
      rbfPointSet = NULL;
    }
    else {
      RendererInterface * renderer = RendererManager::getInstance ()->getMainRenderer ();
      renderer->renderFrame (); // for erasing drawn objects
      throw PrimitiveManagerException ("Incorrect number of points for RBF"); 
    }
  }
}
