#ifndef SPHEREMANAGER_H
#define SPHEREMANAGER_H

#include <qvector.h>
#include <qpainter.h>
#include <Object.h>

#include "../Attributes/ColorAttribute.h"
#include "../Interfaces/PrimitiveManagerInterface.h"
#include "Sphere.h"


/// Events manager for spheres.
/**
 * The SphereManager is able to create spheres, based on the UI events and the 
 * current Scene. A cirle, respresenting the sphere is update during the creation
 * for having a feedback without re-rendering (drawn in 2D).
 */
class SphereManager : public PrimitiveManagerInterface {
public :

  static SphereManager * getInstance ();
  static void destroyInstance ();

  PrimitiveInterface * getCurrentPrimitive (); 
  Object * getCurrentObject ();
  PrimitiveInterface * createPrimitive ();

  /**
   * Gets the <code>SurfelInterface</code> corresponding to the clicked position 
   * of the mouse from <code>FrameBufferInterface</code>, and creates an <code>Sphere</code>
   * with this <code>SurfelInterface</code> as the center of the sphere and radius equal to 0.
   *
   * @param event the UI event gived by CoTeX.
   *
   * @param confWidget the CoTeXConfigurationWidget, 
   *                   for quering options of creation (current texture, etc ...).
   *
   * @see #SurfelInterface
   * @see #FrameBufferInterface
   */
  void handleMousePressEvent (QMouseEvent *event, 
			      CoTeXConfigurationWidget * confWidget);

  /**
   * Handles the event of moving the mouse : update the radius of current sphere.
   * @param event the UI event gived by CoTeX.
   */
  void handleMouseMoveEvent (QMouseEvent *event);
  
  /**
   * Define the final radius (distance beetween original and current position).
   * The current <code>Sphere</code> is ready to be gived by <code>getInstance ()</code>.
   * An Object representing the sphere is aviable (<code>getObject ()</code>).
   *
   * @see #SurfelInterface
   * @see #FrameBufferInterface
   */
  void handleMouseReleaseEvent (QMouseEvent *);
  
  /// Not useful for sphere creation
  void handleKeyPressEvent (QKeyEvent *ke);

  /// Not useful for sphere creation
  void handleKeyReleaseEvent (QKeyEvent *ke);
  
protected :

  /// Protected, because of the design pattern <code>Singleton</code>
  SphereManager ();  
  /// Protected, because of the design pattern <code>Singleton</code> 
  virtual ~SphereManager ();

private :
  
  bool mousePressed;
  QPoint originalMousePosition; ///< the screen original position
  QPoint lastMousePosition; ///< the screen last position
  bool circleVisible;
  float screenRadius;
  Vector3D currentCenter;
  CoTeXConfigurationWidget * configurationWidget;
  
  Sphere * currentPrimitive; ///< the sphere currently constructed
  Object * currentPrimitiveObject; ///< the PS object representing the current instance
  int instanceCount; ///< for generating unique names.
  
  QPainter painter; ///< for painting visual artefacts.
  
  /// Destroy current non null resources. 
  void clean ();

  /// Draw a circle, with the same postion  and radius than the current Sphere.
  /**
   * The circle is based on the screen radius, itself used for computing the real 
   * radius (in the 3D space) of the Sphere. 
   *
   * @param event the mouse event, useful for current mouse position.
   */
  void drawCircle ();
  
  
};

#endif // SPHEREMANAGER_H
