#include <qbuttongroup.h>
#include <qlistbox.h>
#include <qpushbutton.h>
#include <qtabwidget.h>
#include <qlayout.h>
#include <qfiledialog.h>
#include <iostream>
#include <assert.h>

#include "CoTeXConfigurationWidget.h"

#include "TextureConfigurationWidget.h"
#include "../src/Textures/SimpleColorTexture.h"
#include "../src/Textures/PerlinNoiseColorTexture.h"

static const QSize WIDGET_SIZE = QSize (240, 120);
static const QSize PUSHBUTTON_SIZE = QSize (50, 20);
static const QSize LISTBOX_SIZE = QSize (110, 90);

enum PrimitivesIDs {
  SPHERE_ID,
  CONVOLUTION_ID,
  RBF_ID, 
  BLOCK_ID
};

enum OperatorsIDs {
  UNION_ID
};


CoTeXConfigurationWidget::CoTeXConfigurationWidget (TextureInterface::TextureMap * tMap, 
						    QWidget * parent, 
						    const char * name, 
						    WFlags fl) 
  : QWidget (parent, name, fl), texMap (tMap) {
  
  resizeWidget (this, WIDGET_SIZE); 
  setCaption ("CoTeX Tool Settings");
  
  // *************************
  // creation of the tab table 
  // *************************
  
  QTabWidget * CoTeXTab = new QTabWidget (this);
  resizeWidget (CoTeXTab, WIDGET_SIZE);
    
  // redefinition of the policy size
  QFont CoTeXTab_font (CoTeXTab->font ());
  CoTeXTab_font.setPointSize (COTEX_FONT_SIZE);
  CoTeXTab->setFont (CoTeXTab_font); 

  // ************************
  // creation of the CTT page
  // ************************
  
  QWidget * CTTPage = new QWidget (CoTeXTab);

  QButtonGroup * CTTButtonGroup = new QButtonGroup ("", CTTPage);
  resizeWidget (CTTButtonGroup, WIDGET_SIZE); 
  
  QPushButton * CTTLoadPushButton = createPushButton ("load", CTTButtonGroup, PUSHBUTTON_SIZE);
  connect (CTTLoadPushButton, SIGNAL (clicked ()),
	   this, SLOT (handleLoad ()));
  
  QPushButton * CTTSavePushButton = createPushButton ("save", CTTButtonGroup, PUSHBUTTON_SIZE);
  connect (CTTSavePushButton, SIGNAL (clicked ()),
	   this, SLOT (handleSave ()));
  
  QPushButton * CTTExportHFPushButton = createPushButton ("exportHF", CTTButtonGroup, PUSHBUTTON_SIZE);
  connect (CTTExportHFPushButton, SIGNAL (clicked ()),
	   this, SLOT (handleExportHF ()));
  
  QPushButton * CTTResetPushButton = createPushButton ("reset", CTTButtonGroup, PUSHBUTTON_SIZE);
  connect (CTTResetPushButton, SIGNAL (clicked ()), SIGNAL (resetSignal ()));
  
  QGridLayout * CTTLayout = new QGridLayout (CTTButtonGroup, 3, 2, 10, 10);
  CTTLayout->addWidget (CTTLoadPushButton, 0, 0);
  CTTLayout->addWidget (CTTSavePushButton, 0, 1);
  CTTLayout->addWidget (CTTExportHFPushButton, 1, 0);
  CTTLayout->addWidget (CTTResetPushButton, 1, 1);
  CTTLayout->activate ();

  CoTeXTab->insertTab (CTTPage,"CTT");

  // *******************************
  // creation of the primitives page 
  // *******************************

  QWidget * primitivesPage = new QWidget (CoTeXTab);

  primitivesListBox = new QListBox (primitivesPage);
  resizeWidget (primitivesListBox, LISTBOX_SIZE);
  primitivesListBox->insertItem (QPixmap ("Resources/CoTeXSphereIcon.png"), 
				 "sphere", SPHERE_ID);
  primitivesListBox->insertItem (QPixmap ("Resources/CoTeXConvolutionIcon.png"), 
				 "convpoint", CONVOLUTION_ID);
  primitivesListBox->insertItem (QPixmap ("Resources/CoTeXRbfIcon.png"), 
				 "rbf", RBF_ID);
  primitivesListBox->insertItem (QPixmap ("Resources/CoTeXBlockIcon.png"),
				 "block", BLOCK_ID);
  primitivesListBox->setSelected (SPHERE_ID, TRUE);
  

  primitivesWidgetStack = new QWidgetStack (primitivesPage);
  // creation of the elements of primitiveWidgetStack
  convConfigurationWidget = new ConvConfigurationWidget (primitivesWidgetStack);
  rbfConfigurationWidget = new RBFConfigurationWidget (primitivesWidgetStack);
  QWidget * defaultConfigurationWidget = new QWidget (primitivesWidgetStack);
  // update of the configurationWidgetMap
  configurationWidgetMap[CONVOLUTION_ID] = convConfigurationWidget;
  configurationWidgetMap[RBF_ID] = rbfConfigurationWidget;
  configurationWidgetMap[SPHERE_ID] = defaultConfigurationWidget;
  configurationWidgetMap[BLOCK_ID] = defaultConfigurationWidget;
  // addition of the widgets in the primitivesWidgetStack 
  ConfigurationWidgetMap::Iterator it;
  for (it = configurationWidgetMap.begin (); 
       it != configurationWidgetMap.end ();
       it++)
    primitivesWidgetStack->addWidget (it.data (), it.key ());

  primitivesWidgetStack->raiseWidget (defaultConfigurationWidget); // by default
  // commands the display of the widget associate with the current primitive
  connect (primitivesListBox, 
	   SIGNAL (highlighted (int)),
	   this, 
	   SLOT (primitiveDisplay (int)));

  QHBoxLayout * primitivesHBoxlayout = new QHBoxLayout (primitivesPage);
  primitivesHBoxlayout->addWidget (primitivesListBox);
  primitivesHBoxlayout->addWidget (primitivesWidgetStack);
  primitivesHBoxlayout->activate ();

  CoTeXTab->insertTab (primitivesPage, "Primitives");
  
  // ***************************** 
  // creation of the textures page 
  // *****************************

  textureConfigurationWidget = new TextureConfigurationWidget (texMap);
  connect (textureConfigurationWidget,
	   SIGNAL (addNewTextureName (QString, QPixmap)), 
	   this, 
	   SLOT (addNewTextureName (QString, QPixmap)));
  
  QWidget * texturesPage = new QWidget (CoTeXTab);
  
  QPushButton * newTexturePushButton = createPushButton ("new", texturesPage, PUSHBUTTON_SIZE);
  connect (newTexturePushButton, SIGNAL (clicked ()),
	   this, SLOT (appearTextureWidget ()));

  QVBoxLayout * buttonsTextureVBoxLayout = new QVBoxLayout ();
  buttonsTextureVBoxLayout->addWidget (newTexturePushButton);
    
  texturesListBox = new QListBox (texturesPage);
  resizeWidget (texturesListBox, LISTBOX_SIZE); 
  // addition of default textures 
  defaultTextures ();
  texturesListBox->setSelected (0, true);

  QHBoxLayout * texturesHBoxLayout = new QHBoxLayout (texturesPage);
  texturesHBoxLayout->addWidget (texturesListBox);
  texturesHBoxLayout->addLayout (buttonsTextureVBoxLayout);
  texturesHBoxLayout->activate ();
  
  CoTeXTab->insertTab (texturesPage, "Textures");
  
  // ******************************
  // creation of the operators page
  // ******************************

  QWidget * operatorsPage = new QWidget (CoTeXTab);
  
  operatorsListBox = new QListBox (operatorsPage);
  resizeWidget (operatorsListBox, LISTBOX_SIZE);
  operatorsListBox->insertItem (QPixmap ("Resources/CoTeXUnionIcon.png"), 
				"union", UNION_ID);
  operatorsListBox->setSelected (UNION_ID, TRUE);
  
  CoTeXTab->insertTab (operatorsPage, "Operators");
}






void CoTeXConfigurationWidget::resizeWidget (QWidget * widget, QSize size) {
  widget->resize (size);
  widget->setMinimumSize (size);
  widget->setMaximumSize (size);
}

QPushButton * CoTeXConfigurationWidget::createPushButton (QString label,
							  QWidget * parent,
							  QSize size) {
  QPushButton * button = new QPushButton (label, parent);
  resizeWidget(button, size);
  return button;
}

void CoTeXConfigurationWidget::defaultTextures () {
  insertTexture (new PerlinNoiseColorTexture ("R&Vnoise", ColorAttribute (255, 0, 0),
					      ColorAttribute (0, 255, 0)));
  insertTexture (new PerlinNoiseColorTexture ("B&Jnoise", ColorAttribute (0, 0, 255),
					      ColorAttribute (255, 255, 0)));
  insertTexture (new PerlinNoiseColorTexture ("N&Bnoise"));
  insertTexture (new SimpleColorTexture ("yellow", ColorAttribute (255, 255, 0)));
  insertTexture (new SimpleColorTexture ("blue", ColorAttribute (0, 0, 255)));
  insertTexture (new SimpleColorTexture ("green", ColorAttribute (0, 255, 0)));
  insertTexture (new SimpleColorTexture ("red", ColorAttribute (255, 0, 0)));
}

// ********
// Services
// ********
TextureInterface * CoTeXConfigurationWidget::getTexture (QString textureName) const {
  if (texMap->contains (textureName))
    return (*texMap)[textureName];
  return NULL;
}

void CoTeXConfigurationWidget::insertTexture (SimpleColorTexture * tex) {
  QString texName = tex->getName ();
  if (texMap->contains (texName)) {
    delete (*texMap)[texName];
    texMap->remove (texName);
    delete texturesListBox->findItem (texName);
  }
  textureConfigurationWidget->insertTexture (tex);
}

void CoTeXConfigurationWidget::insertTexture (PerlinNoiseColorTexture * tex) {
  QString texName = tex->getName ();
  if (texMap->contains (texName)) {
    delete (*texMap)[texName];
    texMap->remove (texName);
    delete texturesListBox->findItem (texName);
  }
  textureConfigurationWidget->insertTexture (tex);
}


// *****
// SLOTS
// *****

void CoTeXConfigurationWidget::handleLoad () {
  QString filename = QFileDialog::getOpenFileName ("", 
						   "CoTeX files (*.ctx)", 
						   this);
  emit loadSignal (filename);
}

void CoTeXConfigurationWidget::handleSave () {
  QString filename = QFileDialog::getSaveFileName ("newfile.ctx", 
						   "CoTeX files (*.ctx)", 
						   this);
  emit saveSignal (filename);
}

void CoTeXConfigurationWidget::handleExportHF () {
  QString filename = QFileDialog::getSaveFileName ("newfile.hf", 
						   "HyperFun files (*.hf)", 
						   this);
  emit exportHFSignal (filename);
}
  
void CoTeXConfigurationWidget::appearTextureWidget () {
  textureConfigurationWidget->show ();
}

void CoTeXConfigurationWidget::addNewTextureName (QString nameTexture, QPixmap pix) {
  texturesListBox->insertItem (pix, nameTexture, 0); // inserts at the head
  texturesListBox->setSelected (texturesListBox->findItem (nameTexture), TRUE);
  texturesListBox->setTopItem (0); // places the scroll for displaying the item at the top
}

void CoTeXConfigurationWidget::primitiveDisplay (int primitiveId) {
  primitivesWidgetStack->raiseWidget (configurationWidgetMap[primitiveId]);
}  

